"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.$unorderedEqual = exports.$setequal = exports.$equal = exports.$bytesequal = exports.$compare = exports.$bytescompare = exports.$datecompare = exports.$numcompare = void 0;
const commons_1 = require("./commons");
const tsdate_1 = require("./tsdate");
const types_1 = require("./types");
function $numcompare(a, b) {
    if (isNaN(a) || isNaN(b)) {
        return undefined;
    }
    return a === b ? types_1.Same : (a < b ? types_1.Ascending : types_1.Descending);
}
exports.$numcompare = $numcompare;
function $datecompare(a, b) {
    if (!(0, commons_1.$ok)(a) || !(0, commons_1.$ok)(b)) {
        return undefined;
    }
    if (a === b) {
        return types_1.Same;
    }
    if (a instanceof Date && b instanceof Date) {
        return $numcompare(a.getTime(), b.getTime());
    }
    if (!(a instanceof tsdate_1.TSDate)) {
        a = new tsdate_1.TSDate(a);
    }
    if (!(b instanceof tsdate_1.TSDate)) {
        b = new tsdate_1.TSDate(b);
    }
    return a.compare(b);
}
exports.$datecompare = $datecompare;
function $bytescompare(a, b) {
    if (!(0, commons_1.$ok)(a) || !(0, commons_1.$ok)(b)) {
        return undefined;
    }
    if (a === b) {
        return types_1.Same;
    }
    const na = a.length, nb = b.length;
    let i = 0;
    while (i < na && i < nb) {
        const c = $numcompare(a[i], b[i]);
        if (c !== types_1.Same) {
            return c;
        }
        i++;
    }
    return na === nb ? types_1.Same : (i < na ? types_1.Descending : types_1.Ascending);
}
exports.$bytescompare = $bytescompare;
function $compare(a, b) {
    if (!(0, commons_1.$ok)(a) || !(0, commons_1.$ok)(b)) {
        return undefined;
    }
    if (a === b) {
        return types_1.Same;
    }
    if (typeof a === 'number' && typeof b === 'number') {
        return $numcompare(a, b);
    }
    if ((0, commons_1.$isstring)(a) && (0, commons_1.$isstring)(b)) {
        return a > b ? types_1.Descending : (a < b ? types_1.Ascending : types_1.Same);
    }
    if ((0, commons_1.$isarray)(a) && (0, commons_1.$isarray)(b)) {
        const na = a.length, nb = b.length;
        let i = 0;
        while (i < na && i < nb) {
            const c = $compare(a[i], b[i]);
            if (c !== types_1.Same) {
                return c;
            }
            i++;
        }
        return na === nb ? types_1.Same : (i < na ? types_1.Descending : types_1.Ascending);
    }
    if ((a instanceof Date || a instanceof tsdate_1.TSDate) && (b instanceof Date || b instanceof tsdate_1.TSDate)) {
        return $datecompare(a, b);
    }
    if (a instanceof Buffer && b instanceof Buffer) {
        return Buffer.compare(a, b);
    }
    if (a instanceof Uint8Array && b instanceof Uint8Array) {
        return $bytescompare(a, b);
    }
    if ((a instanceof ArrayBuffer || ArrayBuffer.isView(a)) && (b instanceof ArrayBuffer || ArrayBuffer.isView(b))) {
        return $bytescompare(new Uint8Array(a), new Uint8Array(b));
    }
    return (0, commons_1.$isobject)(a) && ('compare' in a) ? a.compare(b) : undefined;
}
exports.$compare = $compare;
function $bytesequal(a, b) {
    if (a === b) {
        return true;
    }
    if (!(0, commons_1.$ok)(a) || !(0, commons_1.$ok)(b))
        return false;
    const n = a.length;
    if (n !== b.length)
        return false;
    for (let i = 0; i < n; i++) {
        if (a[i] !== b[i])
            return false;
    }
    return true;
}
exports.$bytesequal = $bytesequal;
function $equal(a, b) {
    if (a === b) {
        return true;
    }
    if (typeof a === 'number' && typeof b === 'number') {
        return a === b;
    }
    if (!(0, commons_1.$ok)(a) || !(0, commons_1.$ok)(b))
        return false;
    if ((0, commons_1.$isarray)(a) && (0, commons_1.$isarray)(b)) {
        const n = a.length;
        if (n !== b.length)
            return false;
        for (let i = 0; i < n; i++) {
            if (!$equal(a[i], b[i]))
                return false;
        }
        return true;
    }
    if (a instanceof Date && b instanceof Date) {
        return a.getTime() === b.getTime();
    }
    if ((a instanceof Date || a instanceof tsdate_1.TSDate) && (b instanceof Date || b instanceof tsdate_1.TSDate)) {
        if (a instanceof Date) {
            a = new tsdate_1.TSDate(a);
        }
        if (b instanceof Date) {
            b = new tsdate_1.TSDate(b);
        }
        return a.isEqual(b);
    }
    if ((0, commons_1.$isobject)(a) && ('isEqual' in a)) {
        return a.isEqual(b);
    }
    if ((0, commons_1.$isobject)(b) && ('isEqual' in b)) {
        return b.isEqual(a);
    }
    if (a instanceof Set && b instanceof Set) {
        return $setequal(a, b);
    }
    if (a instanceof Map && b instanceof Map) {
        const ak = a.keys();
        const bk = b.keys();
        const keys = a.size >= b.size ? ak : bk;
        for (let k of keys) {
            if (!$equal(a.get(k), b.get(k)))
                return false;
        }
        return true;
    }
    if (a instanceof Buffer && b instanceof Buffer) {
        return Buffer.compare(a, b) === 0;
    }
    if (a instanceof Uint8Array && b instanceof Uint8Array) {
        return $bytesequal(a, b);
    }
    if ((a instanceof ArrayBuffer || ArrayBuffer.isView(a)) && (b instanceof ArrayBuffer || ArrayBuffer.isView(b))) {
        return $bytesequal(new Uint8Array(a), new Uint8Array(b));
    }
    if (Object.getPrototypeOf(a) === Object.prototype && Object.getPrototypeOf(b) === Object.prototype) {
        const ak = Object.getOwnPropertyNames(a);
        const bk = Object.getOwnPropertyNames(a);
        const keys = ak.length >= bk.length ? ak : bk;
        for (let k of keys) {
            if (!$equal(a[k], b[k]))
                return false;
        }
        return true;
    }
    return false;
}
exports.$equal = $equal;
function $setequal(sa, sb) {
    if (sa === sb) {
        return true;
    }
    return (0, commons_1.$ok)(sa) && (0, commons_1.$ok)(sb) && sa.size === sb.size && [...sa.keys()].every(e => sb.has(e));
}
exports.$setequal = $setequal;
function $unorderedEqual(sa, sb) {
    if (sa === sb) {
        return true;
    }
    if (!(0, commons_1.$ok)(sa) || !(0, commons_1.$ok)(sb)) {
        return false;
    }
    let a = undefined;
    let b = undefined;
    let na = -1;
    let nb = -2;
    if ((0, commons_1.$isarray)(sa)) {
        na = (0, commons_1.$count)(sa);
        a = new Set(sa);
    }
    else if (sa instanceof Set) {
        na = sa.size;
        a = sa;
    }
    if ((0, commons_1.$isarray)(sb)) {
        nb = (0, commons_1.$count)(sb);
        b = new Set(sb);
    }
    else if (sb instanceof Set) {
        nb = sb.size;
        b = sb;
    }
    return na === nb && $setequal(a, b);
}
exports.$unorderedEqual = $unorderedEqual;
//# sourceMappingURL=compare.js.map