"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.$copyFile = exports.$realMoveFile = exports.$removeFile = exports.$localRenameFile = exports.$fullWriteBuffer = exports.$writeBuffer = exports.$readData = exports.$readBuffer = exports.$fullWriteString = exports.$writeString = exports.$readString = exports.$loadJSON = exports.$filename = exports.$dir = exports.$newext = exports.$withoutext = exports.$ext = exports.$path = exports.$currentdirectory = exports.$absolute = exports.$normalizepath = exports.$isabsolutepath = exports.$uniquefile = exports.$temporarypath = exports.$filesize = exports.$createDirectory = exports.$isdirectory = exports.$iswritable = exports.$chmod = exports.$isreadable = exports.$isexecutable = exports.$isfile = exports.$stats = void 0;
const fs_1 = require("fs");
const path_1 = require("path");
const os_1 = require("os");
const commons_1 = require("./commons");
const tsdefaults_1 = require("./tsdefaults");
const crypto_1 = require("./crypto");
const utils_1 = require("./utils");
const tsdata_1 = require("./tsdata");
const tserrors_1 = require("./tserrors");
const strings_1 = require("./strings");
const tscharset_1 = require("./tscharset");
const data_1 = require("./data");
const isWindows = typeof process === "object" && process.platform === "win32";
function $stats(src) {
    tserrors_1.TSError.assertNotInBrowser('$stats');
    if ((0, commons_1.$length)(src)) {
        try {
            return (0, fs_1.statSync)(src);
        }
        catch (e) {
            if ((e === null || e === void 0 ? void 0 : e.code) === 'ENOENT')
                return null;
            throw e;
        }
    }
    return null;
}
exports.$stats = $stats;
function $isfile(src) {
    tserrors_1.TSError.assertNotInBrowser('$isfile');
    const stats = $stats(src);
    return (0, commons_1.$ok)(stats) ? stats.isFile() : false;
}
exports.$isfile = $isfile;
function $isexecutable(src) {
    tserrors_1.TSError.assertNotInBrowser('$isexecutable');
    return $isfile(src) && _safeCheckPermissions(src, fs_1.constants.R_OK | fs_1.constants.X_OK);
}
exports.$isexecutable = $isexecutable;
function $isreadable(src) {
    tserrors_1.TSError.assertNotInBrowser('$isreadable');
    return (0, commons_1.$ok)($stats(src)) && _safeCheckPermissions(src, fs_1.constants.R_OK);
}
exports.$isreadable = $isreadable;
function $chmod(src, mode) {
    tserrors_1.TSError.assertNotInBrowser('$chmod');
    let ret = false;
    if ((0, commons_1.$length)(src) && (0, commons_1.$isunsigned)(mode) && mode <= 0o777) {
        try {
            (0, fs_1.chmodSync)(src, mode);
            ret = true;
        }
        catch (_a) {
            ret = false;
        }
    }
    return ret;
}
exports.$chmod = $chmod;
function $iswritable(src) {
    tserrors_1.TSError.assertNotInBrowser('$iswritable');
    return (0, commons_1.$ok)($stats(src)) && _safeCheckPermissions(src, fs_1.constants.W_OK);
}
exports.$iswritable = $iswritable;
function $isdirectory(src) {
    tserrors_1.TSError.assertNotInBrowser('$isdirectory');
    const stats = $stats(src);
    return (0, commons_1.$ok)(stats) ? stats.isDirectory() : false;
}
exports.$isdirectory = $isdirectory;
function $createDirectory(p) {
    tserrors_1.TSError.assertNotInBrowser('$createDirectory');
    let ret = false;
    if ((0, commons_1.$length)(p)) {
        const stats = $stats(p);
        if ((0, commons_1.$ok)(stats) && stats.isDirectory()) {
            ret = true;
        }
        else if (!(0, commons_1.$ok)(stats) || !stats.isFile()) {
            try {
                (0, fs_1.mkdirSync)(p, { recursive: true });
                ret = true;
            }
            catch (_a) {
                ret = false;
            }
        }
    }
    return ret;
}
exports.$createDirectory = $createDirectory;
function $filesize(src) {
    tserrors_1.TSError.assertNotInBrowser('$filesize');
    const stats = $stats(src);
    return (0, commons_1.$ok)(stats) ? stats.size : 0;
}
exports.$filesize = $filesize;
function $temporarypath(ext = '', src = '') {
    tserrors_1.TSError.assertNotInBrowser('$temporarypath');
    ext = (0, strings_1.$ftrim)(ext);
    let file = $uniquefile(src);
    if ((0, commons_1.$length)(ext)) {
        file = $newext(file, ext);
    }
    return $path((0, tsdefaults_1.$tmp)(), file);
}
exports.$temporarypath = $temporarypath;
function $uniquefile(src, e = undefined, internalImplementation = false) {
    const rand = (0, crypto_1.$uuid)();
    e = (0, strings_1.$ftrim)(e);
    if (!(0, commons_1.$length)(src)) {
        return $newext(rand, e, internalImplementation);
    }
    const finalExt = e.length ? e : $ext(src);
    return $newext(`${$withoutext(src, internalImplementation)}-${rand}`, finalExt, internalImplementation);
}
exports.$uniquefile = $uniquefile;
function $isabsolutepath(src, internalImplementation = false) {
    const browser = (0, utils_1.$inbrowser)();
    return (0, commons_1.$length)(src) > 0 ? (!isWindows && (internalImplementation || browser) ? src.startsWith('/') : (0, path_1.isAbsolute)(src)) : false;
}
exports.$isabsolutepath = $isabsolutepath;
function $normalizepath(src, internalImplementation = false) {
    if (!(0, commons_1.$length)(src)) {
        return '.';
    }
    return !isWindows && (internalImplementation || (0, utils_1.$inbrowser)()) ? $path(true, src) : (0, path_1.normalize)(src);
}
exports.$normalizepath = $normalizepath;
function $absolute(src, internalImplementation = false) {
    tserrors_1.TSError.assertNotInBrowser('$absolute');
    if (!(0, commons_1.$length)(src)) {
        return $currentdirectory();
    }
    if ($isabsolutepath(src)) {
        return src;
    }
    if (src.startsWith('~')) {
        return $path(internalImplementation, (0, os_1.homedir)(), src.slice(1));
    }
    return $path(internalImplementation, $currentdirectory(), src);
}
exports.$absolute = $absolute;
function $currentdirectory() {
    tserrors_1.TSError.assertNotInBrowser('$currentdirectory');
    return (0, path_1.resolve)(process.cwd());
}
exports.$currentdirectory = $currentdirectory;
function $path(first, ...paths) {
    const browser = (0, utils_1.$inbrowser)();
    function _internalPath(...paths) {
        const n = paths.length;
        if (n > 0) {
            const firstComponents = paths[0].split('/');
            const isAbsolute = paths[0].startsWith('/');
            const comps = [];
            function _addComponent(comps, s) {
                if (s === '..') {
                    if (comps.length > 0) {
                        if (comps[comps.length - 1] != '..') {
                            comps.pop();
                        }
                        else {
                            comps.push('..');
                        }
                    }
                    else if (!isAbsolute) {
                        comps.push('..');
                    }
                }
                else if (s.length > 0 && s !== '.') {
                    comps.push(s);
                }
            }
            for (let p of firstComponents) {
                _addComponent(comps, p);
            }
            for (let i = 1; i < n; i++) {
                const p = paths[i];
                if (p.length > 0) {
                    for (let s of p.split('/')) {
                        _addComponent(comps, s);
                    }
                }
            }
            if (comps.length) {
                const s = comps.join('/');
                return isAbsolute ? '/' + s : s;
            }
            else if (isAbsolute) {
                return '/';
            }
        }
        return '';
    }
    if (!(0, commons_1.$isstring)(first)) {
        return (!isWindows && first || browser) ? _internalPath(...paths) : (0, path_1.join)(...paths);
    }
    return !isWindows && browser ? _internalPath(first, ...paths) : (0, path_1.join)(first, ...paths);
}
exports.$path = $path;
function $ext(s, internalImplementation = false) {
    if ((0, commons_1.$length)(s)) {
        if (!isWindows && (internalImplementation || (0, utils_1.$inbrowser)())) {
            const slash = s.lastIndexOf('/');
            const dot = s.lastIndexOf('.');
            return dot > slash + 1 ? s.slice(dot + 1) : '';
        }
        else {
            const e = (0, path_1.extname)(s);
            if ((0, commons_1.$length)(e)) {
                return e.slice(1);
            }
        }
    }
    return '';
}
exports.$ext = $ext;
function $withoutext(s, internalImplementation) {
    if (!(0, commons_1.$length)(s)) {
        return '';
    }
    const e = $ext(s, internalImplementation);
    s = e.length ? s.slice(0, s.length - e.length) : s;
    return s.length && s.charAt(s.length - 1) === '.' ? s.slice(0, s.length - 1) : s;
}
exports.$withoutext = $withoutext;
function $newext(s, e = undefined, internalImplementation) {
    let b = $withoutext(s, internalImplementation);
    return (0, commons_1.$length)(e) ? `${b}.${e}` : b;
}
exports.$newext = $newext;
function $dir(source, internalImplementation = false) {
    let len = (0, commons_1.$length)(source);
    if (!len) {
        return '.';
    }
    let s = source;
    if (!isWindows && (internalImplementation || (0, utils_1.$inbrowser)())) {
        let start = 0;
        while (start < len && s.charAt(start) === '/') {
            start++;
        }
        if (start === len) {
            return '/';
        }
        else if (len === 1) {
            return '.';
        }
        let end = len - 1;
        while (end > start && s.charAt(end) === '/') {
            end--;
        }
        ;
        end++;
        if (start === 0 && end === len) {
            const slash = s.lastIndexOf('/');
            return slash > 0 ? s.slice(0, slash) : '.';
        }
        else {
            const t = s.slice(start, end);
            const slash = t.lastIndexOf('/');
            return slash > 0 ? s.slice(0, start) + t.slice(0, slash) : (start > 0 ? s.slice(0, start) : '.');
        }
    }
    return (0, path_1.dirname)(s);
}
exports.$dir = $dir;
function $filename(s, internalImplementation = false) {
    let len = (0, commons_1.$length)(s);
    if (!len) {
        return '';
    }
    if (!isWindows && (internalImplementation || (0, utils_1.$inbrowser)())) {
        let end = len - 1;
        while (end >= 0 && s.charAt(end) === '/') {
            end--;
        }
        ;
        end++;
        if (end === 0) {
            return '';
        }
        else if (end < len) {
            s = s.slice(0, end);
        }
        const slash = s.lastIndexOf('/');
        return slash >= 0 ? s.slice(slash + 1) : s;
    }
    return (0, path_1.basename)(s);
}
exports.$filename = $filename;
function $loadJSON(source) {
    const pathParameter = (0, commons_1.$isstring)(source);
    if (pathParameter) {
        tserrors_1.TSError.assertNotInBrowser('$loadJSON');
    }
    let ret = null;
    const src = source instanceof ArrayBuffer ? (0, data_1.$bufferFromArrayBuffer)(source) : source;
    if ((0, commons_1.$length)(src)) {
        const charset = tscharset_1.TSCharset.utf8Charset();
        let loadedString = pathParameter ? $readString(src, charset) : charset.stringFromData(source);
        if ((0, commons_1.$length)(loadedString)) {
            try {
                ret = JSON.parse(loadedString);
                ret = (0, commons_1.$ok)(ret) ? ret : null;
            }
            catch (e) {
                if (pathParameter) {
                    (0, utils_1.$logterm)(`&R&w Impossible to parse JSON file &P ${src} &0`);
                }
                else {
                    (0, utils_1.$logterm)(`&R&w Impossible to parse &P ${src.length} bytes &R given JSON buffer &0`);
                }
                ret = null;
            }
        }
    }
    return ret;
}
exports.$loadJSON = $loadJSON;
function $readString(src, encoding) {
    tserrors_1.TSError.assertNotInBrowser('$readString');
    const buffer = $readBuffer(src);
    return (0, commons_1.$ok)(buffer) ? (0, tscharset_1.$charset)(encoding).stringFromData(buffer) : null;
}
exports.$readString = $readString;
function $writeString(src, str, opts = {}) {
    tserrors_1.TSError.assertNotInBrowser('$writeString');
    const [res,] = $fullWriteString(src, str, opts);
    return res;
}
exports.$writeString = $writeString;
function $fullWriteString(src, str, opts) {
    tserrors_1.TSError.assertNotInBrowser('$fullWriteString');
    return $fullWriteBuffer(src, (0, tscharset_1.$charset)(opts.encoding).dataFromString(str, opts.stringStart, opts.stringEnd), opts);
}
exports.$fullWriteString = $fullWriteString;
function $readBuffer(src) {
    tserrors_1.TSError.assertNotInBrowser('$readBuffer');
    let ret = null;
    if ((0, commons_1.$length)(src)) {
        try {
            ret = (0, fs_1.readFileSync)(src);
        }
        catch (e) {
            ret = null;
        }
    }
    return ret;
}
exports.$readBuffer = $readBuffer;
function $readData(src) {
    tserrors_1.TSError.assertNotInBrowser('$readData');
    const buf = $readBuffer(src);
    return (0, commons_1.$ok)(buf) ? new tsdata_1.TSData(buf, { dontCopySourceBuffer: true }) : null;
}
exports.$readData = $readData;
function $writeBuffer(src, buf, opts = {}) {
    tserrors_1.TSError.assertNotInBrowser('$writeBuffer');
    const [res,] = $fullWriteBuffer(src, buf, opts);
    return res;
}
exports.$writeBuffer = $writeBuffer;
function $fullWriteBuffer(src, buf, opts) {
    tserrors_1.TSError.assertNotInBrowser('$fullWriteBuffer');
    let done = false;
    let precedent = null;
    let start = (0, commons_1.$ok)(opts.byteStart) ? opts.byteStart : 0;
    let end = (0, commons_1.$ok)(opts.byteEnd) ? opts.byteEnd : buf.byteLength;
    let mode = (0, commons_1.$ok)(opts.mode) ? opts.mode : 0o666;
    if ((0, commons_1.$length)(src) && (0, commons_1.$isunsigned)(start) && (0, commons_1.$isunsigned)(end) && (0, commons_1.$isunsigned)(mode)) {
        const pathToWrite = opts.attomically ? $uniquefile(src) : src;
        end = Math.min(end, buf.byteLength);
        start = Math.min(start, end);
        if (buf instanceof tsdata_1.TSData) {
            [buf,] = buf.internalStorage;
        }
        try {
            const fd = (0, fs_1.openSync)(pathToWrite, 'w', mode);
            const MAX_TRY = 3;
            try {
                let retries = 0;
                while (start < end) {
                    const wlen = (0, fs_1.writeSync)(fd, buf, start, end - start);
                    if (wlen === 0) {
                        retries++;
                    }
                    else {
                        retries = 0;
                    }
                    if (retries > MAX_TRY) {
                        throw new tserrors_1.TSError(`$fullWriteBuffer(): tried to fs.writeSync() ${retries} times without any success.`, {
                            path: src,
                            buffer: buf,
                            options: opts
                        });
                    }
                    start += wlen;
                }
                done = true;
            }
            finally {
                (0, fs_1.closeSync)(fd);
            }
        }
        catch (e) {
            done = false;
        }
        if (done && opts.attomically) {
            const renamedExistingFile = $uniquefile(src);
            done = _safeRename(src, renamedExistingFile);
            if (done && !_safeRename(pathToWrite, src)) {
                if (!_safeRename(renamedExistingFile, src)) {
                    throw new tserrors_1.TSError(`Unable to atomically finish writing file '${src}'`, {
                        wantedPath: src,
                        renamedExistingFile: renamedExistingFile,
                        writtenDataFile: pathToWrite
                    });
                }
                done = false;
            }
            if (!done) {
                _safeUnlink(pathToWrite);
            }
            else if (opts.removePrecedentVersion) {
                _safeUnlink(renamedExistingFile);
            }
            else {
                precedent = renamedExistingFile;
            }
        }
    }
    return [done, precedent];
}
exports.$fullWriteBuffer = $fullWriteBuffer;
function $localRenameFile(src, dest) {
    tserrors_1.TSError.assertNotInBrowser('$localRenameFile');
    return (0, commons_1.$length)(src) > 0 && (0, commons_1.$length)(dest) > 0 && src !== dest && $isfile(src) && _safeRename(src, dest);
}
exports.$localRenameFile = $localRenameFile;
function $removeFile(src) {
    tserrors_1.TSError.assertNotInBrowser('$removeFile');
    return (0, commons_1.$length)(src) > 0 && $isfile(src) && _safeUnlink(src);
}
exports.$removeFile = $removeFile;
function $realMoveFile(src, dest) {
    tserrors_1.TSError.assertNotInBrowser('$realMoveFile');
    let done = false;
    if ((0, commons_1.$length)(src) && (0, commons_1.$length)(dest) && src !== dest && $isfile(src)) {
        if ($isdirectory(dest)) {
            dest = $path(dest, $filename(src));
            if (src === dest) {
                return false;
            }
        }
        done = _safeRename(src, dest);
        if (!done) {
            try {
                (0, fs_1.copyFileSync)(src, dest);
                (0, fs_1.unlinkSync)(src);
                done = true;
            }
            catch (_a) {
                done = false;
            }
        }
    }
    return done;
}
exports.$realMoveFile = $realMoveFile;
function $copyFile(src, dest, overwrite = false) {
    tserrors_1.TSError.assertNotInBrowser('$copyFile');
    let done = false;
    if ((0, commons_1.$length)(src) && (0, commons_1.$length)(dest) && src !== dest && $isfile(src)) {
        if ($isdirectory(dest)) {
            dest = $path(dest, $filename(src));
        }
        if (src === dest || $isdirectory(dest) || (!overwrite && $isfile(dest))) {
            return false;
        }
        try {
            (0, fs_1.copyFileSync)(src, dest, (overwrite ? 0 : fs_1.constants.COPYFILE_EXCL));
            done = true;
        }
        catch (_a) {
            done = false;
        }
    }
    return done;
}
exports.$copyFile = $copyFile;
function _safeRename(s, d) { let done = false; try {
    (0, fs_1.renameSync)(s, d);
    done = true;
}
catch (_a) {
    done = false;
} return done; }
function _safeUnlink(p) { let done = false; try {
    (0, fs_1.unlinkSync)(p);
    done = true;
}
catch (_a) {
    done = false;
} return done; }
function _safeCheckPermissions(src, permissions) {
    let ret = false;
    try {
        if ((0, fs_1.statSync)(src).isFile()) {
            (0, fs_1.accessSync)(src, permissions);
            ret = true;
        }
    }
    catch (_a) {
        ret = false;
    }
    return ret;
}
//# sourceMappingURL=fs.js.map